import { Ref, ComputedRef } from 'vue-demi';
import { ResizeObserver } from '@juggle/resize-observer';
import color2k from 'color2k';

declare type Fn = () => void;

declare type RefElement = Element | Ref<Element | undefined>;
declare type MaybeRef<T> = T | Ref<T>;

/**
 * Size of a specific box.
 *
 * https://drafts.csswg.org/resize-observer-1/#resizeobserversize
 */
interface ResizeObserverSize {
    readonly inlineSize: number;
    readonly blockSize: number;
}
/**
 * Used to observe different box sizes.
 *
 * https://drafts.csswg.org/resize-observer-1/#enumdef-resizeobserverboxoptions
 */
declare enum ResizeObserverBoxOptions {
    BORDER_BOX = "border-box",
    CONTENT_BOX = "content-box",
    DEVICE_PIXEL_CONTENT_BOX = "device-pixel-content-box"
}
/**
 * Options to be given to the resize observer,
 * when observing a new element.
 *
 * https://drafts.csswg.org/resize-observer-1/#dictdef-resizeobserveroptions
 */
interface ResizeObserverOptions {
    box?: 'content-box' | 'border-box' | 'device-pixel-content-box' | ResizeObserverBoxOptions;
}
/**
 * https://drafts.csswg.org/resize-observer-1/#resize-observer-entry-interface
 */
interface ResizeObserverEntry {
    readonly target: Element;
    readonly contentRect: DOMRectReadOnly;
    readonly borderBoxSize?: ReadonlyArray<ResizeObserverSize>;
    readonly contentBoxSize?: ReadonlyArray<ResizeObserverSize>;
    readonly devicePixelContentBoxSize?: ReadonlyArray<ResizeObserverSize>;
}
/**
 * https://drafts.csswg.org/resize-observer-1/#resize-observer-callback
 */
declare type ResizeObserverCallback = (entries: ResizeObserverEntry[], observer: ResizeObserver) => void;

interface WindowSize {
    width: Ref<number>;
    height: Ref<number>;
}
interface ElementSize extends WindowSize {
}
interface EventOptions {
    useCapture?: boolean;
    useThrottle?: boolean | null;
    delay?: number;
}

interface ReactiveColor {
    hexColor: ComputedRef<string>;
    hslaColor: ComputedRef<string>;
    rgbaColor: ComputedRef<string>;
    readableColor: ComputedRef<string>;
    invertColor: ComputedRef<string>;
    luminance: ComputedRef<number>;
    lightenColor: ComputedRef<string>;
    darkenColor: ComputedRef<string>;
    tintColor: ComputedRef<string>;
    shadeColor: ComputedRef<string>;
    desaturateColor: ComputedRef<string>;
    saturateColor: ComputedRef<string>;
    opacifyColor: ComputedRef<string>;
    transparentizeColor: ComputedRef<string>;
}

declare const useCssVars: (el: MaybeRef<HTMLElement | null>, variable: string, defaultValue: string) => Ref<string>;

/**
 * enhanced `color2k` with new utilities like `tint` 、 `shade` 、 `invert`.
 */
declare const usePalette: {
    adjustHue: typeof color2k.adjustHue;
    darken: typeof color2k.darken;
    desaturate: typeof color2k.desaturate;
    getContrast: typeof color2k.getContrast;
    getLuminance: typeof color2k.getLuminance;
    getScale: typeof color2k.getScale;
    guard: typeof color2k.guard;
    hasBadContrast: typeof color2k.hasBadContrast;
    hsla: typeof color2k.hsla;
    lighten: typeof color2k.lighten;
    mix: typeof color2k.mix;
    opacify: typeof color2k.opacify;
    parseToHsla: typeof color2k.parseToHsla;
    readableColor: typeof color2k.readableColor;
    readableColorIsBlack: typeof color2k.readableColorIsBlack;
    rgba: typeof color2k.rgba;
    saturate: typeof color2k.saturate;
    transparentize: typeof color2k.transparentize;
    parseToRgba: typeof color2k.parseToRgba;
    ColorError: typeof color2k.ColorError;
    toHex: typeof color2k.toHex;
    toRgba: typeof color2k.toRgba;
    toHsla: typeof color2k.toHsla;
    /**
     * Tints a color by mixing it with white.
     * @param color
     * @param amount given as a decimal between 0 and 1
     */
    tint: (color: string, amount: number) => string;
    /**
     * Shades a color by mixing it with black.
     * @param color
     * @param amount given as a decimal between 0 and 1
     */
    shade: (color: string, amount: number) => string;
    /**
     * Inverts the red, green and blue values of a color.
     * @param color
     */
    invert: (color: string) => string;
};
/**
 * Reactive color parsing and manipulation built on top of color2k.
 * @param color
 * @param amount
 */
declare const useColor: (color: Ref<string>, amount?: Ref<number> | undefined) => ReactiveColor;

declare const useDebouncedRef: (value: any, delay?: number) => Ref;

declare const useElementSize: (target: Element, options: ResizeObserverOptions) => ElementSize;

declare function useEventListener<E extends keyof WindowEventMap>(target: Window, event: E, listener: (this: Window, ev: WindowEventMap[E]) => any, options?: boolean | AddEventListenerOptions): Fn;
declare function useEventListener<E extends keyof DocumentEventMap>(target: Document, event: E, listener: (this: Document, ev: DocumentEventMap[E]) => any, options?: boolean | AddEventListenerOptions): Fn;
declare function useEventListener<E extends keyof DocumentEventMap>(target: RefElement | Ref<any>, event: E, listener: (this: Element, ev: DocumentEventMap[E]) => any, options?: boolean | AddEventListenerOptions): Fn;
declare function useEventListener(target: MaybeRef<EventTarget>, event: string, listener: EventListener, options?: boolean | AddEventListenerOptions): Fn;

interface IFullScreen {
    isFullscreen: Ref<boolean>;
    toggleFullscreen: Fn;
}
/**
 * Reactive Fullscreen API build on top of screenfull.js
 * @param {HTMLElement} target
 * @return {boolean} isFullscreen
 * @return {function} toggleFullscreen
 */
declare const useFullscreen: (target: HTMLElement) => IFullScreen;

declare const useInstance: () => any;

/**
 * Call onMounted() if it's inside a component lifecycle, if not, run just call the function
 * @param fn
 * @param sync if set to false, it will run in the nextTick() of Vue
 */
declare function tryOnMounted(fn: Fn, sync?: boolean): void;
/**
 * Call onUnmounted() if it's inside a component lifecycle, if not, do nothing
 * @param fn
 */
declare function tryOnUnmounted(fn: Fn): void;

/**
 * Register `MouseMove` event on mounted, and remove event automatically on unmounted.
 */
declare const useMouse: () => {
    x: Ref<number>;
    y: Ref<number>;
    remove: () => void;
};

declare function useReactiveState(getter: () => Required<object>): [Required<object>, VoidFunction];

/**
 * Reactive Resize API, When window resize, do someting.
 * @param {function} fn
 * @param {object} options
 * @return {function} dispatchResize
 */
declare const useResizeEvent: (fn: Fn, options?: EventOptions) => {
    dispatchResize: () => void;
};

/**
 * A map of keybinding strings to event handlers.
 */
interface KeyBindingMap {
    [keybinding: string]: (event: KeyboardEvent) => void;
}
interface KeymasterEvent {
    key: string;
    method: KeyHandler;
    mods: number[];
    scope: string;
    shortcut: string;
}
interface KeyHandler {
    (keyboardEvent: KeyboardEvent, keymasterEvent: KeymasterEvent): void;
}
/**
 * Bind shortcut onMounted & unbind shortcut automatically onUnmounted in setup hooks
 * or
 * Unbind shortcut before bind shortcut without setup hooks
 * @param {Object} keymap
 * @param {Object} vm
 * @return {function} toggle
 */
declare const useShortcut: (keymap: KeyBindingMap, vm: any) => {
    shift: boolean;
    alt: boolean;
    option: boolean;
    ctrl: boolean;
    control: boolean;
    command: boolean;
    setScope(scopeName: string): void;
    getScope(): string;
    deleteScope(scopeName: string): void;
    noConflict(): void;
    unbind(key: string): void;
    unbind(key: string, scopeName: string): void;
    isPressed(keyCode: number): boolean;
    getPressedKeyCodes(): number[];
    filter(event: FilterEvent): void;
    bind: (seed: string, func: Function) => void;
};

/**
 * It immediately detects when an element resizes and provides accurate sizing information back to the handler.
 * @param target
 * @param callback
 * @param options
 */
declare const useResizeObserver: (target: Element, callback: ResizeObserverCallback, options: ResizeObserverOptions) => {
    stop: () => void;
};

/**
 * Reactive window size
 * @param {object} options
 * @return {Ref<number>} WindowSize
 */
declare const useWindowSize: (options: EventOptions) => WindowSize;

export { IFullScreen, KeyBindingMap, KeyHandler, KeymasterEvent, tryOnMounted, tryOnUnmounted, useColor, useCssVars, useDebouncedRef, useElementSize, useEventListener, useFullscreen, useInstance, useMouse, usePalette, useReactiveState, useResizeEvent, useResizeObserver, useShortcut, useWindowSize };
