/*!
 * @vueblocks/vue-use-core v0.2.4
 * (c) 2021 xiaoluoboding
 * @license MIT
 */
import { ref, watch, computed, isRef, unref, customRef, getCurrentInstance, onMounted, nextTick, onUnmounted, isVue3, reactive } from 'vue-demi';
import color2k from 'color2k';
import debounce from 'lodash.debounce';
import { ResizeObserver } from '@juggle/resize-observer';
import screenfull from 'screenfull';
import throttle from 'lodash.throttle';
import keymaster from 'keymaster';

const cssVariableRegex = /--[\S]*/g;
const useCssVars = (el, variable, defaultValue) => {
    if (!variable || !variable.match(cssVariableRegex)) {
        console.warn(`[useCssVars] css variable is not valid: ${variable}`);
    }
    const propertyValue = ref('');
    const target = ref(el || window.document.documentElement);
    watch(target, (el) => {
        if (window && el) {
            const propValue = window.getComputedStyle(el).getPropertyValue(variable) || defaultValue || '';
            propertyValue.value = propValue;
        }
    }, { immediate: true });
    watch(propertyValue, (newVal) => {
        if (target.value && target.value.style) {
            target.value.style.setProperty(variable, newVal);
        }
    });
    return propertyValue;
};

const isBoolean = (val) => typeof val === 'boolean';
const isClient = ![typeof window, typeof document].includes('undefined');
const isElement = (val) => typeof val === 'object' && !!val.tagName;
const isUndefined = (val) => val === undefined;

const NOOP = () => { };

/**
 * enhanced `color2k` with new utilities like `tint` 、 `shade` 、 `invert`.
 */
const usePalette = {
    /**
     * Tints a color by mixing it with white.
     * @param color
     * @param amount given as a decimal between 0 and 1
     */
    tint: (color, amount) => color2k.mix('#FFF', color, amount),
    /**
     * Shades a color by mixing it with black.
     * @param color
     * @param amount given as a decimal between 0 and 1
     */
    shade: (color, amount) => color2k.mix('#000', color, amount),
    /**
     * Inverts the red, green and blue values of a color.
     * @param color
     */
    invert: (color) => {
        if (color === 'transparent')
            return color;
        const [r, g, b, a] = color2k.parseToRgba(color);
        return color2k.rgba(255 - r, 255 - g, 255 - b, a);
    },
    ...color2k
};
/**
 * Reactive color parsing and manipulation built on top of color2k.
 * @param color
 * @param amount
 */
const useColor = (color, amount) => {
    const { tint, shade, invert } = usePalette;
    const colorState = {
        hexColor: computed(() => ''),
        hslaColor: computed(() => ''),
        rgbaColor: computed(() => ''),
        readableColor: computed(() => ''),
        invertColor: computed(() => ''),
        luminance: computed(() => 0),
        lightenColor: computed(() => ''),
        darkenColor: computed(() => ''),
        tintColor: computed(() => ''),
        shadeColor: computed(() => ''),
        desaturateColor: computed(() => ''),
        saturateColor: computed(() => ''),
        opacifyColor: computed(() => ''),
        transparentizeColor: computed(() => '')
    };
    // accpet color argument
    if (isRef(color) && unref(color)) {
        colorState.hexColor = computed(() => color2k.toHex(unref(color)));
        colorState.hslaColor = computed(() => color2k.toHsla(unref(color)));
        colorState.rgbaColor = computed(() => color2k.toRgba(unref(color)));
        colorState.readableColor = computed(() => color2k.readableColor(unref(color)));
        colorState.invertColor = computed(() => invert(unref(color)));
        colorState.luminance = computed(() => color2k.getLuminance(unref(color)));
    }
    // accept amount argument
    if (isRef(amount) && !isUndefined(unref(amount))) {
        colorState.lightenColor = computed(() => color2k.lighten(unref(color), unref(amount)));
        colorState.darkenColor = computed(() => color2k.darken(unref(color), unref(amount)));
        colorState.tintColor = computed(() => tint(unref(color), unref(amount)));
        colorState.shadeColor = computed(() => shade(unref(color), unref(amount)));
        colorState.desaturateColor = computed(() => color2k.desaturate(unref(color), unref(amount)));
        colorState.saturateColor = computed(() => color2k.saturate(unref(color), unref(amount)));
        colorState.opacifyColor = computed(() => color2k.opacify(unref(color), unref(amount)));
        colorState.transparentizeColor = computed(() => color2k.transparentize(unref(color), unref(amount)));
    }
    return colorState;
};

const useDebouncedRef = (value, delay = 200) => {
    return customRef((track, trigger) => {
        return {
            get() {
                track();
                return value;
            },
            set: debounce((newValue) => {
                value = newValue;
                trigger();
            }, delay)
        };
    });
};

/**
 * It immediately detects when an element resizes and provides accurate sizing information back to the handler.
 * @param target
 * @param callback
 * @param options
 */
const useResizeObserver = (target, callback, options) => {
    let ro;
    const targetRef = ref(target);
    const isSupported = window && 'ResizeObserver' in window;
    const stopWatch = watch(targetRef, (newVal) => {
        if (isSupported && window && newVal) {
            ro = new ResizeObserver(callback);
            ro.observe(newVal, options);
        }
    }, { immediate: true });
    const stop = () => {
        ro && ro.disconnect();
        ro = undefined;
        stopWatch();
    };
    return {
        stop
    };
};

const useElementSize = (target, options) => {
    const width = ref(isElement(target) ? target.clientWidth : 0);
    const height = ref(isElement(target) ? target.clientHeight : 0);
    useResizeObserver(target, ([entry]) => {
        width.value = entry.contentRect.width;
        height.value = entry.contentRect.height;
    }, options);
    return { width, height };
};

function useEventListener(target, event, listener, options) {
    let remove = NOOP;
    if (target) {
        const element = ref(target);
        const removeEventListener = (e) => e.removeEventListener(event, listener);
        const addEventListener = (e) => e.addEventListener(event, listener, options);
        const removeWatch = watch(element, (n, _, cleanUp) => {
            if (n) {
                addEventListener(n);
                cleanUp(() => removeEventListener(n));
            }
        }, { immediate: true });
        remove = () => {
            removeEventListener(element.value);
            removeWatch();
        };
    }
    return remove;
}
// export { useEventListener }

/**
 * Call onMounted() if it's inside a component lifecycle, if not, run just call the function
 * @param fn
 * @param sync if set to false, it will run in the nextTick() of Vue
 */
function tryOnMounted(fn, sync = true) {
    if (getCurrentInstance()) {
        onMounted(fn);
    }
    else if (sync) {
        fn();
    }
    else {
        nextTick(fn);
    }
}
/**
 * Call onUnmounted() if it's inside a component lifecycle, if not, do nothing
 * @param fn
 */
function tryOnUnmounted(fn) {
    if (getCurrentInstance())
        onUnmounted(fn);
}

/**
 * Reactive Fullscreen API build on top of screenfull.js
 * @param {HTMLElement} target
 * @return {boolean} isFullscreen
 * @return {function} toggleFullscreen
 */
const useFullscreen = (target) => {
    const targetRef = ref(target || document.querySelector('html'));
    const isFullscreen = ref(false);
    const toggleFullscreen = () => {
        screenfull.isEnabled && screenfull.toggle(targetRef.value);
    };
    tryOnMounted(() => {
        if (screenfull.isEnabled) {
            screenfull.on('change', () => {
                isFullscreen.value = screenfull.isFullscreen;
            });
        }
    });
    return {
        isFullscreen,
        toggleFullscreen
    };
};

const useInstance = () => {
    const instance = getCurrentInstance();
    if (!instance) {
        console.error('`getCurrentInstance` only works during setup or Lifecycle Hooks');
        return null;
    }
    return isVue3 ? instance.ctx : instance.proxy;
};

/**
 * Register `MouseMove` event on mounted, and remove event automatically on unmounted.
 */
const useMouse = () => {
    let remove = NOOP;
    const x = ref(0);
    const y = ref(0);
    // const touchHandler = (event: TouchEvent) => {
    //   if (event.touches.length > 0) {
    //     x.value = event.touches[0].clientX
    //     y.value = event.touches[0].clientY
    //   }
    // }
    if (window) {
        const mouseHandler = (event) => {
            x.value = event.pageX;
            y.value = event.pageY;
        };
        remove = useEventListener(window, 'mousemove', mouseHandler, { passive: true });
    }
    return {
        x,
        y,
        remove
    };
};

function useReactiveState(getter) {
    const state = reactive(getter());
    const update = () => {
        Object.assign(state, getter());
    };
    return [state, update];
}

/**
 * Reactive Resize API, When window resize, do someting.
 * @param {function} fn
 * @param {object} options
 * @return {function} dispatchResize
 */
const useResizeEvent = (fn, options = { useCapture: false, useThrottle: null, delay: 500 }) => {
    const { useCapture = false, useThrottle = null, delay = 500 } = options;
    const dispatchResize = () => {
        const e = document.createEvent('Event');
        e.initEvent('resize', true, true);
        window.dispatchEvent(e);
    };
    const waitFn = isBoolean(useThrottle)
        ? useThrottle
            ? throttle(fn, delay)
            : debounce(fn, delay)
        : fn;
    tryOnMounted(() => {
        window.addEventListener('resize', waitFn, useCapture);
    });
    tryOnUnmounted(() => {
        window.removeEventListener('resize', waitFn, useCapture);
    });
    return {
        dispatchResize
    };
};

const bindKeyHandler = (fn) => {
    return (...args) => {
        fn(...args);
        return false;
    };
};
/**
 * Given shortcut a new bind function
 */
const shortcut = {
    bind: (seed, func) => keymaster(seed, bindKeyHandler(func)),
    ...keymaster
};
/**
 * Bind shortcut onMounted & unbind shortcut automatically onUnmounted in setup hooks
 * or
 * Unbind shortcut before bind shortcut without setup hooks
 * @param {Object} keymap
 * @param {Object} vm
 * @return {function} toggle
 */
const useShortcut = (keymap, vm) => {
    let sync = true;
    if (!vm) {
        vm = getCurrentInstance();
    }
    else {
        sync = false;
        unbindKeys();
    }
    function bindKeys() {
        Object.keys(keymap).forEach(key => {
            shortcut.bind(key, keymap[key]);
        });
    }
    function unbindKeys() {
        Object.keys(keymap).forEach(key => {
            shortcut.unbind(key);
        });
    }
    tryOnMounted(() => bindKeys(), sync);
    tryOnUnmounted(unbindKeys);
    return shortcut;
};

/**
 * Reactive window size
 * @param {object} options
 * @return {Ref<number>} WindowSize
 */
const useWindowSize = (options) => {
    const width = ref(isClient ? window.innerWidth : 0);
    const height = ref(isClient ? window.innerHeight : 0);
    const onResize = () => {
        width.value = window.innerWidth;
        height.value = window.innerHeight;
    };
    const { dispatchResize } = useResizeEvent(onResize, options);
    onMounted(dispatchResize);
    return { width, height };
};

export { tryOnMounted, tryOnUnmounted, useColor, useCssVars, useDebouncedRef, useElementSize, useEventListener, useFullscreen, useInstance, useMouse, usePalette, useReactiveState, useResizeEvent, useResizeObserver, useShortcut, useWindowSize };
